;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "acp" "0.13.1"
  "An ACP (Agent Client Protocol) implementation."
  '((emacs "28.1"))
  :url "https://github.com/xenodium/acp.el"
  :commit "a29cb161ac95f1819f34481a98666707661c5cf8"
  :revdesc "a29cb161ac95")
